@@ -7,19 +7,27 @@ from scipy.sparse import lil_matrix
 
 
 class OpenCLContext:
-    def __init__(self, p, a, w, d, e, r, c, q,i, j, k, l):
-        self.MAX_GPU_PROC_ELEM = p
-        self.IS_OPENCL_AVAILABLE = a
-        self.work_group_size = w
+    def __init__(self, a, b, g, d, e, f, c, q, i, j, k, l, m, n, o, p, x, y, z, w):
+        self.MAX_GPU_PROC_ELEM = a
+        self.IS_OPENCL_AVAILABLE = b
+        self.work_group_size = g
         self.dtype = d
         self.epsilon = e
-        self.max_alloc = r
+        self.max_alloc = f
         self.ctx = c
         self.queue = q
         self.calculate_cut_kernel = i
         self.calculate_cut_sparse_kernel = j
         self.calculate_cut_segmented_kernel = k
         self.calculate_cut_sparse_segmented_kernel = l
+        self.single_bit_flips_kernel = m
+        self.single_bit_flips_sparse_kernel = n
+        self.single_bit_flips_segmented_kernel = o
+        self.single_bit_flips_sparse_segmented_kernel = p
+        self.double_bit_flips_kernel = x
+        self.double_bit_flips_sparse_kernel = y
+        self.double_bit_flips_segmented_kernel = z
+        self.double_bit_flips_sparse_segmented_kernel = w
 
 IS_OPENCL_AVAILABLE = True
 ctx = None
@@ -33,6 +41,14 @@ calculate_cut_kernel = None
 calculate_cut_sparse_kernel = None
 calculate_cut_segmented_kernel = None
 calculate_cut_sparse_segmented_kernel = None
+single_bit_flips_kernel = None
+single_bit_flips_sparse_kernel = None
+single_bit_flips_segmented_kernel = None
+single_bit_flips_sparse_segmented_kernel = None
+double_bit_flips_kernel = None
+double_bit_flips_sparse_kernel = None
+double_bit_flips_segmented_kernel = None
+double_bit_flips_sparse_segmented_kernel = None
 
 dtype_bits = int(os.getenv('PYQRACKISING_FPPOW', '5'))
 kernel_src = ''
@@ -85,6 +101,14 @@ try:
     calculate_cut_sparse_kernel = program.calculate_cut_sparse
     calculate_cut_segmented_kernel = program.calculate_cut_segmented
     calculate_cut_sparse_segmented_kernel = program.calculate_cut_sparse_segmented
+    single_bit_flips_kernel = program.single_bit_flips
+    single_bit_flips_sparse_kernel = program.single_bit_flips_sparse
+    single_bit_flips_segmented_kernel = program.single_bit_flips_segmented
+    single_bit_flips_sparse_segmented_kernel = program.single_bit_flips_sparse_segmented
+    double_bit_flips_kernel = program.double_bit_flips
+    double_bit_flips_sparse_kernel = program.double_bit_flips_sparse
+    double_bit_flips_segmented_kernel = program.double_bit_flips_segmented
+    double_bit_flips_sparse_segmented_kernel = program.double_bit_flips_sparse_segmented
 
     work_group_size = calculate_cut_kernel.get_work_group_info(
         cl.kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE,
@@ -96,7 +120,7 @@ except ImportError:
     IS_OPENCL_AVAILABLE = False
     print("PyOpenCL not installed. (If you have any OpenCL accelerator devices with available ICDs, you might want to optionally install pyopencl.)")
 
-opencl_context = OpenCLContext(compute_units, IS_OPENCL_AVAILABLE, work_group_size, dtype, epsilon, max_alloc, ctx, queue, calculate_cut_kernel, calculate_cut_sparse_kernel, calculate_cut_segmented_kernel, calculate_cut_sparse_segmented_kernel)
+opencl_context = OpenCLContext(compute_units, IS_OPENCL_AVAILABLE, work_group_size, dtype, epsilon, max_alloc, ctx, queue, calculate_cut_kernel, calculate_cut_sparse_kernel, calculate_cut_segmented_kernel, calculate_cut_sparse_segmented_kernel, single_bit_flips_kernel, single_bit_flips_sparse_kernel, single_bit_flips_segmented_kernel, single_bit_flips_sparse_segmented_kernel, double_bit_flips_kernel, double_bit_flips_sparse_kernel, double_bit_flips_segmented_kernel, double_bit_flips_sparse_segmented_kernel)
 
 
 def setup_opencl(l, g, args_np):
@@ -189,6 +213,20 @@ def make_theta_buf(theta, is_segmented, shots, n):
     return theta_buf
 
 
+def make_best_theta_buf(theta):
+    n = theta.shape[0]
+    n32 = (n + 31) >> 5
+    theta_np = np.zeros(n32, dtype=np.uint32)
+    for i in range(n):
+        if theta[i]:
+            theta_np[(i >> 5)] |= 1 << (i & 31)
+
+    mf = cl.mem_flags
+    theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=theta_np)
+
+    return theta_buf
+
+
 @njit(parallel=True)
 def convert_bool_to_uint(samples):
     shots = samples.shape[0]
@@ -373,7 +411,7 @@ def init_thresholds(n_qubits):
 
 
 @njit
-def probability_by_hamming_weight(J, h, z, theta, t, n_bias):
+def probability_by_hamming_weight(J, h, z, theta, t, n_bias, normalized=True):
     zJ = z * J
     theta_c = ((np.pi if J > 0 else -np.pi) / 2) if abs(zJ) < epsilon else np.arcsin(max(-1.0, min(1.0, h / zJ)))
 
@@ -397,6 +435,9 @@ def probability_by_hamming_weight(J, h, z, theta, t, n_bias):
     if J > 0.0:
         return bias[::-1]
 
+    if normalized:
+        bias /= bias.sum()
+
     return bias
 
 
@@ -421,8 +462,8 @@ def maxcut_hamming_cdf(n_qubits, J_func, degrees, quality, tot_t, h_mult):
         t = step * delta_t
         tm1 = (step - 1) * delta_t
         h_t = h_mult * (tot_t - t)
-        bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, t, n_bias)
-        last_bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, tm1, n_bias)
+        bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, t, n_bias, False)
+        last_bias = probability_by_hamming_weight(J_eff, h_t, z, theta_eff, tm1, n_bias, False)
         for i in range(n_bias):
             hamming_prob[i] += bias[i] - last_bias[i]
 
@@ -483,22 +524,22 @@ def bit_pick(weights, used, n):
 
 
 @njit
-def gray_code_next(state, curr_idx):
+def gray_code_next(state, curr_idx, offset):
     prev = curr_idx
     curr = curr_idx + 1
     prev = prev ^ (prev >> 1)
     curr = curr ^ (curr >> 1)
     diff = prev ^ curr
     flip_bit = int(np.log2(diff))
-    state[flip_bit] = not state[flip_bit]
+    state[offset + flip_bit] = not state[offset + flip_bit]
 
 
 @njit
-def gray_mutation(index, seed_bits):
+def gray_mutation(index, seed_bits, offset):
     """Apply Gray-code-indexed bit flips to a seed bitstring."""
     n = seed_bits.shape[0]
     gray = index ^ (index >> 1)
     bits = seed_bits.copy()
     for i in range(n):
-        bits[n - 1 - i] ^= (gray >> i) & 1
+        bits[n - offset - 1 - i] ^= (gray >> i) & 1
     return bits
