@@ -1,11 +1,9 @@
 from .maxcut_tfim_streaming import maxcut_tfim_streaming
 from .maxcut_tfim_util import compute_cut_streaming, compute_energy_streaming, get_cut, get_cut_base, gray_code_next, gray_mutation, int_to_bitstring, opencl_context
-import itertools
 import networkx as nx
 import numpy as np
 from numba import njit, prange
 import os
-import random
 
 
 dtype = opencl_context.dtype
@@ -37,33 +35,58 @@ def run_single_bit_flips(best_theta, is_spin_glass, G_func, nodes):
 
 
 @njit(parallel=True)
-def run_double_bit_flips(best_theta, is_spin_glass, G_func, nodes, combos):
+def run_double_bit_flips(best_theta, is_spin_glass, G_func, nodes, thread_count):
     n = len(best_theta)
-    combo_count = (n * (n - 1)) >> 1
+    combo_count = (n * (n - 1)) // 2
+    thread_batch = (combo_count + thread_count - 1) // thread_count
 
-    states = np.empty((combo_count, n), dtype=np.bool_)
-    energies = np.empty(combo_count, dtype=dtype)
+    states = np.empty((thread_count, n), dtype=np.bool_)
+    energies = np.empty(thread_count, dtype=dtype)
 
     if is_spin_glass:
-        for c in prange(combo_count):
-            state = best_theta.copy()
-            c2 = c << 1
-            b = combos[c2]
-            state[b] = not state[b]
-            b = combos[c2 + 1]
-            state[b] = not state[b]
+        for t in prange(thread_count):
+            s = int(t)
+            while s < combo_count:
+                c = int(s)
+                i = int(0)
+                lcv = int(n - 1)
+                while c >= lcv:
+                    c -= lcv
+                    i += 1
+                    lcv -= 1
+                    if not lcv:
+                        break
+                j = int(c + i + 1)
 
-            states[c], energies[c] = state, compute_energy_streaming(state, G_func, nodes, n)
+                state = best_theta.copy()
+                state[i] = not state[i]
+                state[j] = not state[j]
+
+                states[t], energies[t] = state, compute_energy_streaming(state, G_func, nodes, n)
+
+                s += thread_batch
     else:
-        for c in prange(combo_count):
-            state = best_theta.copy()
-            c2 = c << 1
-            b = combos[c2]
-            state[b] = not state[b]
-            b = combos[c2 + 1]
-            state[b] = not state[b]
+        for t in prange(thread_count):
+            s = int(t)
+            while s < combo_count:
+                c = int(s)
+                i = int(0)
+                lcv = int(n - 1)
+                while c >= lcv:
+                    c -= lcv
+                    i += 1
+                    lcv -= 1
+                    if not lcv:
+                        break
+                j = int(c + i + 1)
 
-            states[c], energies[c] = state, compute_cut_streaming(state, G_func, nodes, n)
+                state = best_theta.copy()
+                state[i] = not state[i]
+                state[j] = not state[j]
+
+                states[t], energies[t] = state, compute_cut_streaming(state, G_func, nodes, n)
+
+                s += thread_batch
 
     best_index = np.argmax(energies)
     best_energy = energies[best_index]
@@ -74,21 +97,27 @@ def run_double_bit_flips(best_theta, is_spin_glass, G_func, nodes, combos):
 
 @njit(parallel=True)
 def pick_gray_seeds(best_theta, thread_count, gray_seed_multiple, G_func, nodes, n, is_spin_glass):
-    seed_count = thread_count * gray_seed_multiple
+    blocks = (n + 63) // 64
+    block_size = thread_count * gray_seed_multiple
+    seed_count = block_size * blocks
 
     seeds = np.empty((seed_count, n), dtype=np.bool_)
     energies = np.empty(seed_count, dtype=dtype)
 
     if is_spin_glass:
-        for i in prange(seed_count):
-            seed = gray_mutation(i, best_theta)
-            energies[i] = compute_energy_streaming(seed, G_func, nodes, n)
-            seeds[i] = seed
+        for s in prange(seed_count):
+            i = s % block_size
+            offset = (s // block_size) * 64
+            seed = gray_mutation(i, best_theta, offset)
+            energies[s] = compute_energy_streaming(seed, G_func, nodes, n)
+            seeds[s] = seed
     else:
-        for i in prange(seed_count):
-            seed = gray_mutation(i, best_theta)
-            energies[i] = compute_cut_streaming(seed, G_func, nodes, n)
-            seeds[i] = seed
+        for s in prange(seed_count):
+            i = s % block_size
+            offset = (s // block_size) * 64
+            seed = gray_mutation(i, best_theta, offset)
+            energies[s] = compute_cut_streaming(seed, G_func, nodes, n)
+            seeds[s] = seed
 
     indices = np.argsort(energies)[::-1]
 
@@ -103,30 +132,41 @@ def pick_gray_seeds(best_theta, thread_count, gray_seed_multiple, G_func, nodes,
     return best_seeds, best_energies
 
 @njit(parallel=True)
-def run_gray_optimization(best_theta, iterators, gray_iterations, thread_count, is_spin_glass, G_func, nodes):
+def run_gray_optimization(best_theta, iterators, energies, gray_iterations, thread_count, is_spin_glass, G_func, nodes):
     n = len(best_theta)
     thread_iterations = (gray_iterations + thread_count - 1) // thread_count
+    blocks = (n + 63) // 64
 
     states = np.empty((thread_count, n), dtype=np.bool_)
-    energies = np.full(thread_count, np.finfo(dtype).min, dtype=dtype)
 
     if is_spin_glass:
         for i in prange(thread_count):
-            iterator = iterators[i]
+            iterator = iterators[i].copy()
+            best_energy, best_iterator = energies[i], iterator.copy()
             for curr_idx in range(thread_iterations):
-                gray_code_next(iterator, curr_idx)
-                energy = compute_energy_streaming(iterator, G_func, nodes, n)
-                if energy > energies[i]:
-                    states[i], energies[i] = iterator.copy(), energy
+                for block in range(blocks):
+                    gray_code_next(iterator, curr_idx, block * 64)
+                    energy = compute_energy_streaming(iterator, G_func, nodes, n)
+                    if energy > best_energy:
+                        best_iterator, best_energy = iterator.copy(), energy
+                    else:
+                        iterator = best_iterator.copy()
+                if best_energy > energies[i]:
+                    states[i], energies[i] = best_iterator.copy(), best_energy
     else:
         for i in prange(thread_count):
-            iterator = iterators[i]
-            r_iterator = iterator.copy()
+            iterator = iterators[i].copy()
+            best_energy, best_iterator = energies[i], iterator.copy()
             for curr_idx in range(thread_iterations):
-                gray_code_next(iterator, curr_idx)
-                energy = compute_cut_streaming(iterator, G_func, nodes, n)
-                if energy > energies[i]:
-                    states[i], energies[i] = iterator.copy(), energy
+                for block in range(blocks):
+                    gray_code_next(iterator, curr_idx, block * 64)
+                    energy = compute_cut_streaming(iterator, G_func, nodes, n)
+                    if energy > best_energy:
+                        best_iterator, best_energy = iterator.copy(), energy
+                    else:
+                        iterator = best_iterator.copy()
+                if best_energy > energies[i]:
+                    states[i], energies[i] = best_iterator.copy(), best_energy
 
     best_index = np.argmax(energies)
     best_energy = energies[best_index]
@@ -186,8 +226,6 @@ def spin_glass_solver_streaming(
 
     max_energy = compute_energy_streaming(best_theta, G_func, nodes, n_qubits) if is_spin_glass else cut_value
 
-    combos2 = np.array(list(itertools.combinations(range(n_qubits), 2))).flatten()
-
     thread_count = os.cpu_count() ** 2
     improved = True
     while improved:
@@ -202,7 +240,7 @@ def spin_glass_solver_streaming(
             continue
 
         # Double bit flips with O(n^3)
-        energy, state = run_double_bit_flips(best_theta, is_spin_glass, G_func, nodes, combos2)
+        energy, state = run_double_bit_flips(best_theta, is_spin_glass, G_func, nodes, thread_count)
         if energy > max_energy:
             max_energy = energy
             best_theta = state
@@ -218,7 +256,26 @@ def spin_glass_solver_streaming(
             improved = True
             continue
 
-        energy, state = run_gray_optimization(best_theta, iterators, gray_iterations, thread_count, is_spin_glass, G_func, nodes)
+        energy, state = run_gray_optimization(best_theta, iterators, energies, gray_iterations, thread_count, is_spin_glass, G_func, nodes)
+        if energy > max_energy:
+            max_energy = energy
+            best_theta = state
+            improved = True
+            continue
+
+        # Post-reheat phase
+        reheat_theta = state
+
+        # Single bit flips with O(n^2)
+        energy, state = run_single_bit_flips(reheat_theta, is_spin_glass, G_func, nodes)
+        if energy > max_energy:
+            max_energy = energy
+            best_theta = state
+            improved = True
+            continue
+
+        # Double bit flips with O(n^3)
+        energy, state = run_double_bit_flips(reheat_theta, is_spin_glass, G_func, nodes, thread_count)
         if energy > max_energy:
             max_energy = energy
             best_theta = state
